!+
! Module: bkgtemplategen_mod
!
! Authors:  Ian Stewart - JBO
!          
! Description: Uses simple arithmetic to make a background-map template.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module bkgtemplategen_mod

  use dal
  use param  
  use errorHandling

  use dal_aux
  use save_image
  use wcs_aux

  implicit none 

  private 

  ! Command-line parameters:
  character(256) ::&
    expMapSetName='',&
    nonVigSetName='',&
    bkgMapSetName='',&
    tempImageName=''
  real(single) ::&
    vigBkgFlux=0.0,&
    nonVigBkgFlux=0.0

  public bkgtemplategen

contains 

  !---------------------------------------------------------------------
  subroutine bkgtemplategen
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

    call readParameters

    call makeBkgMap(expMapSetName, nonVigSetName, bkgMapSetName&
      , tempImageName, vigBkgFlux, nonVigBkgFlux)

  end subroutine bkgtemplategen

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

    expMapSetName = stringParameter('expmapset')
    nonVigSetName = stringParameter('nonvigset')
    bkgMapSetName = stringParameter('bkgmapset')
    tempImageName = stringParameter('tempset')
    vigBkgFlux    = realParameter('vigbkgrate')
    nonVigBkgFlux = realParameter('nonvigrate')

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine makeBkgMap(expMapSetName, nonVigSetName, bkgMapSetName&
  , tempImageName, vigBkgFlux, nonVigBkgFlux)
  !---------------------------------------------------------------------
  ! This essentially just adds the nonvignetted and vignetted exposure
  ! maps in the specified proportions. It is only done here by a dedicated
  ! subroutine rather than simply by use of farith because it is desirable
  ! to be a little careful in calculating the required amplitudes from
  ! the input vigBkgFlux and nonVigBkgFlux values.

    implicit none

    character(*),   intent(in) :: expMapSetName,&
                                  nonVigSetName,&
                                  bkgMapSetName,&
                                  tempImageName
    real(single),   intent(in) :: vigBkgFlux, nonVigBkgFlux
    ! both in counts / arcsec^2 / s.

    real(single), pointer ::&
      expMapImage(:,:), nonVigImage(:,:)
    real(single) :: pixelAreaArcsec
    type(DataSetT) :: bkgSet
    integer(int16) :: xSize, ySize
    real(single), allocatable :: bkgCountsMapImage(:,:)
    type(WcsT) :: wcs

    !-------------------------------------------------------------------
    ! Start

    call readArrayData(expMapSetName, expMapImage) ! in ssclib/dal_aux.
    call readArrayData(nonVigSetName, nonVigImage) ! in ssclib/dal_aux.

    xSize = size(expMapImage, 1)
    ySize = size(expMapImage, 2)
    if ( size(nonVigImage, 1) .ne. xSize&
    .or. size(nonVigImage, 2) .ne. ySize)&
call error('imageSizesDifferent','The exposure map and &
&the nonvignetted map have different image sizes.')

    allocate(bkgCountsMapImage(xSize, ySize))

    call readWcs(expMapSetName, wcs) ! in ssclib/wcs_aux.

    select case(wcs%x%worldUnit)
    case('deg')
      pixelAreaArcsec = 3600.0 * 3600.0 * wcsPixelAreaWorld(wcs)
      ! wcsPixelAreaWorld is in ssclb/wcs_aux.
    case default
call error('notSupported',"At present, WCS units (CUNITn keywords) must be 'deg'.")
    end select

    bkgCountsMapImage = ((nonVigImage / maxval(nonVigImage))&
      * nonVigBkgFlux * pixelAreaArcsec&
      + (expMapImage / maxval(expMapImage))&
      * vigBkgFlux * pixelAreaArcsec) * maxval(expMapImage)
    ! Units of the bkgCountsMapImage are counts/pixel.

    call saveImage(bkgCountsMapImage, expMapSetName, bkgMapSetName&
      , tempImageName) ! in ssclib/save_image.

    bkgSet = dataSet(bkgMapSetName, MODIFY)
    if (vigBkgFlux + nonVigBkgFlux > 0.0) then
      call setAttribute(bkgSet, 'VIG_FRAC', vigBkgFlux &
        / (vigBkgFlux + nonVigBkgFlux)&
        , comment='Fraction of the total background which is vignetted.')
    end if
    call release(bkgSet)

  end subroutine makeBkgMap

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module bkgtemplategen_mod
