!+
! Module: imstats_mod
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Does the same job as fimgstats, but hopefully without the drawbacks.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module imstats_mod

  use dal
  use param  
  use errorHandling

!  use array_utils	! for valueAtHistoFraction
  use dal_aux		! for readArrayData
!  use math_utils	! for incompleteGammaP
!  use minimizations	! for simplexMinimization, riddersMethod
!  use save_image
!  use wcs_aux		! for readWcs, writeWcs

  implicit none 

  private 

  ! Command-line parameters:
  !
  character(256)::&
    inImageSetName=''
  real(single) ::&
    zLo=0.0,&
    zHi=0.0
  integer ::&
    xiLo=0,&
    xiHi=0,&
    yiLo=0,&
    yiHi=0
  logical ::&
    withZLimits = .false.,&
    withXYLimits = .false.

  public imstats

contains 

  !---------------------------------------------------------------------
  subroutine imstats
  !---------------------------------------------------------------------
!********* quick and dirty for the moment.

    implicit none

    real(double), pointer :: real64Image(:,:)
    real(double) :: av, total, numReal, ssr, rms

    !-------------------------------------------------------------------
    ! Start

    call readParameters

!!******* get imDataType
!
!    select case(imDataType)
!    case(INTEGER8, INTEGER16, INTEGER32)
!!*** load int32 image
!    case(REAL32, REAL64)
!!*** load real64 image
!    case(STRING, BOOLEAN)
!call error('',"")
!    case default
!call error('',"")
!    end select

call readArrayData(inImageSetName, real64Image)

numReal = real(size(real64Image,1),double) * real(size(real64Image,2),double)
total = sum(real64Image)
av = total / numReal
ssr = sum((real64Image-av)*(real64Image-av)) / numReal
rms = sqrt(ssr)

print *, '****** statistics for image ******'
print *, ''
print *, 'The sum of the selected image                = ', total
print *, 'The mean of the selected image               = ', av
print *, 'The standard deviation of the selected image = ', rms
print *, 'The number of points used in calculation     = ', numReal
print *, ''
print *, '****** successfully exited ******'

deallocate(real64Image)

  end subroutine imstats

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

    inImageSetName = stringParameter('inset')
    withZLimits = booleanParameter('withzlimits')
    if (withZLimits) then
      zLo = realParameter('zlo')
      zHi = realParameter('zhi')
    end if
    withXYLimits = booleanParameter('withxylimits')
    if (withXYLimits) then
      xiLo = intParameter('xlo')
      xiHi = intParameter('xhi')
      yiLo = intParameter('ylo')
      yiHi = intParameter('yhi')
    end if

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module imstats_mod
