!
! Module: l_array_utils_2
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains some routines for dealing with keywords and columns of FITS files.
! 
! Comments: TODO:
!*********** Look through ALL software for unnecessary pointer allocations!
!*********** Look through ALL software and declare all ptrs => null()!
! 
! $Id: l_array_utils_2.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- fred
module l_array_utils_2

  use dal
  use errorHandling

  use constants	! PiOn2
  use array_utils
  use sort_mod
!use coordinate
use dump

  implicit none 

  private 

! array_utils:
  public l_getBinNumberSingle
  public l_getBinNumberDouble
  public maxLoc1d
  public minLoc1d

  interface maxLoc1d
    module procedure maxLoc1dSingle, maxLoc1dDouble
  end interface
  interface minLoc1d
    module procedure minLoc1dSingle, minLoc1dDouble
  end interface

  contains

  !---------------------------------------------------------------------
  function maxLoc1dSingle(vector, mask)
  !---------------------------------------------------------------------
  ! I wrote this because I was sick of the mucking around needed to use
  ! maxloc when the argument is a simple vector.
  !.....................................................................

    implicit none

    real(single), intent(in)           :: vector(:)
    logical,      intent(in), optional :: mask(size(vector))
    integer                            :: maxLoc1dSingle

    integer :: maxLocVec(1)

    !-------------------------------------------------------------------
    ! Start

    if (present(mask)) then
      maxLocVec = maxloc(vector, mask)
    else
      maxLocVec = maxloc(vector)
    end if
    maxLoc1dSingle = maxLocVec(1)

  end function maxLoc1dSingle

  !---------------------------------------------------------------------
  function maxLoc1dDouble(vector, mask)
  !---------------------------------------------------------------------
  ! I wrote this because I was sick of the mucking around needed to use
  ! maxloc when the argument is a simple vector.
  !.....................................................................

    implicit none

    real(double), intent(in)           :: vector(:)
    logical,      intent(in), optional :: mask(size(vector))
    integer                            :: maxLoc1dDouble

    integer :: maxLocVec(1)

    !-------------------------------------------------------------------
    ! Start

    if (present(mask)) then
      maxLocVec = maxloc(vector, mask)
    else
      maxLocVec = maxloc(vector)
    end if
    maxLoc1dDouble = maxLocVec(1)

  end function maxLoc1dDouble

  !---------------------------------------------------------------------
  function minLoc1dSingle(vector, mask)
  !---------------------------------------------------------------------
  ! I wrote this because I was sick of the mucking around needed to use
  ! minloc when the argument is a simple vector.
  !.....................................................................

    implicit none

    real(single), intent(in)           :: vector(:)
    logical,      intent(in), optional :: mask(size(vector))
    integer                            :: minLoc1dSingle

    integer :: minLocVec(1)

    !-------------------------------------------------------------------
    ! Start

    if (present(mask)) then
      minLocVec = minloc(vector, mask)
    else
      minLocVec = minloc(vector)
    end if
    minLoc1dSingle = minLocVec(1)

  end function minLoc1dSingle

  !---------------------------------------------------------------------
  function minLoc1dDouble(vector, mask)
  !---------------------------------------------------------------------
  ! I wrote this because I was sick of the mucking around needed to use
  ! minloc when the argument is a simple vector.
  !.....................................................................

    implicit none

    real(double), intent(in)           :: vector(:)
    logical,      intent(in), optional :: mask(size(vector))
    integer                            :: minLoc1dDouble

    integer :: minLocVec(1)

    !-------------------------------------------------------------------
    ! Start

    if (present(mask)) then
      minLocVec = minloc(vector, mask)
    else
      minLocVec = minloc(vector)
    end if
    minLoc1dDouble = minLocVec(1)

  end function minLoc1dDouble


  !---------------------------------------------------------------------
  function l_getBinNumberSingle(xBinEdges, x) result(binNum)
  !---------------------------------------------------------------------
!*** replace the present array_utils sub by this. IMS 2006-05-19.

  ! This function is designed to return the index or number of the bin
  ! (the first bin is binNum = 1) in which the input value x falls. The
  ! function performs as follows:
  !
  ! It compares the value of x with the bin 'edges' stored in xBinEdges.
  !
  ! If x < xBinEdges(1), binNum = 0 is returned. It is up to the calling
  ! routine to catch these instances if it is necessary to prevent this
  ! return value from occurring.
  !
  ! If xBinEdges(i) <= x < xBinEdges(i+1), binNum = i is returned.
  !
  ! If xBinEdges(size(xBinEdges)-1) <= x <= xBinEdges(size(xBinEdges)),
  ! binNum = size(xBinEdges)-1 is returned.
  !
  ! If x > xBinEdges(size(xBinEdges)), binNum = size(xBinEdges) is returned.
  ! It is up to the calling routine to catch these instances if it is
  ! necessary to prevent this return value from occurring.
  !.....................................................................

    implicit none

    real(single), intent(in) :: xBinEdges(:), x
    integer(int32)           :: binNum

    !-------------------------------------------------------------------
    ! Start

    if (size(xBinEdges) < 2) then
      if (size(xBinEdges)==1) then
        if (x > xBinEdges(1)) then
          binNum = 1
        else
          binNum = 0
        end if
      else
        binNum = 0
      end if
  return
    end if

    binNum = largestTrueIndex(xBinEdges<=x)

  end function l_getBinNumberSingle

  !---------------------------------------------------------------------
  function l_getBinNumberDouble(xBinEdges, x) result(binNum)
  !---------------------------------------------------------------------
!*** replace the present array_utils sub by this. IMS 2006-05-19.

  ! This function is designed to return the index or number of the bin
  ! (the first bin is binNum = 1) in which the input value x falls. The
  ! function performs as follows:
  !
  ! It compares the value of x with the bin 'edges' stored in xBinEdges.
  !
  ! If x < xBinEdges(1), binNum = 0 is returned. It is up to the calling
  ! routine to catch these instances if it is necessary to prevent this
  ! return value from occurring.
  !
  ! If xBinEdges(i) <= x < xBinEdges(i+1), binNum = i is returned.
  !
  ! If xBinEdges(size(xBinEdges)-1) <= x <= xBinEdges(size(xBinEdges)),
  ! binNum = size(xBinEdges)-1 is returned.
  !
  ! If x > xBinEdges(size(xBinEdges)), binNum = size(xBinEdges) is returned.
  ! It is up to the calling routine to catch these instances if it is
  ! necessary to prevent this return value from occurring.
  !.....................................................................

    implicit none

    real(double), intent(in) :: xBinEdges(:), x
    integer(int32)           :: binNum

    !-------------------------------------------------------------------
    ! Start

    if (size(xBinEdges) < 2) then
      if (size(xBinEdges)==1) then
        if (x > xBinEdges(1)) then
          binNum = 1
        else
          binNum = 0
        end if
      else
        binNum = 0
      end if
  return
    end if

    binNum = largestTrueIndex(xBinEdges<=x)

  end function l_getBinNumberDouble

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module l_array_utils_2
