!
! Module: random_aux
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains routines to generate random numbers with a variety of probability distributions.
! 
! Comments: TODO:
! 
! $Id: random_aux.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module l_random_aux

  use types

  use constants
  use array_utils	! for getBinNumber()
  use random		! in utils. for SAS_rand.
  use sort_mod
  use ssc_misc		! for fracVal.
!use temp_ssclib	! for decideOption
use l_ssc_misc_2	! for decideOption.

  implicit none 

  private 

  public randomLetterStr
  public segmentRandom
  public getSegmentRandoms
  public poissonRandom
  public getPoissonRandoms
  public minus5on2Random
  public getMinus5on2Randoms
  public powerRandom
  public getPowerRandoms
  public getLogNLogSRandoms
  private doLogNLogSChecks
  public getPowerSegmentRandoms
  public doDiffPowerSegmentChecks
  public doIntPowerSegmentChecks
  public getSineRandoms
  private calcK
  public exponentialRandom
  public getExponentialRandoms
  public gaussianRandom
  public getGaussianRandoms
  public randomNumber
  public getRandomNumbers
  public randoms
  public makeRandomSequence
  public coinIsHeads
  public getRandomOrder

  interface getPoissonRandoms
    module procedure getPoissonRandomsVector
    module procedure getPoissonRandomsArray
  end interface

contains 

  !---------------------------------------------------------------------
  function randomLetterStr(inStr) result(outStr)
  !---------------------------------------------------------------------
  ! Returns a random string of lower-case letters which has the same length as the input string.

    implicit none

    character(*),         intent(in) :: inStr
    character(len(inStr))            :: outStr

    integer :: charI(len(inStr)), startLettersI, finisLettersI, charNum
    real(single) :: numbers(len(inStr))

    !-------------------------------------------------------------------
    ! Start:

    startLettersI = iachar('a')
    finisLettersI = iachar('z')
    call getRandomNumbers(numbers)
    charI = floor(fracVal(real(startLettersI, single)&
      , real(finisLettersI+1, single), numbers))
    where(charI < startLettersI) charI = startLettersI
    where(charI > finisLettersI) charI = finisLettersI
    do charNum = 1, len(inStr)
      outStr(charNum:charNum) = achar(charI(charNum))
    end do

  end function randomLetterStr

  !---------------------------------------------------------------------
  function segmentRandom(x, y)
  !---------------------------------------------------------------------
  ! Returns a random value with a distribution given by the sequence of
  ! line segments defined by vectors x amd y.
  !
  ! Note that x and y should be of size 2 or more, the x values should be
  ! in increasing order and none of the y values should be negative.

    implicit none

    real(single), intent(in)  :: x(:), y(size(x))
    real(single) :: segmentRandom

    integer, parameter :: hsize = 1000
    real(single) :: harvest(hsize)
    integer :: i
    save i, harvest
    data i /1/

    !-------------------------------------------------------------------
    ! Start:

    if (i .eq. 1) then
      call getSegmentRandoms(x, y, harvest)
    end if
    segmentRandom = harvest(i)
    i = i + 1
    if (i > hsize) i = 1

  end function segmentRandom

  !---------------------------------------------------------------------
  subroutine getSegmentRandoms(x, y, numbers)
  !---------------------------------------------------------------------
  ! Returns a vector of random values which have a distribution given by
  ! the sequence of line segments defined by vectors x amd y.
  !
  ! Note that x and y should be of size 2 or more, the x values should be
  ! in increasing order and none of the y values should be negative.

!*** NOT YET TESTED!!!

    implicit none

    real(single), intent(in)  :: x(:), y(size(x))
    real(single), intent(out) :: numbers(:)

    real(single), allocatable :: harvest(:)
    integer :: n, number, numSegments, i, bin
!    logical :: firstCall
!    save firstCall
!    data firstCall /.true./
    real(single), allocatable :: yy(:), integratedSequence(:), curvature(:)
    real(single) :: maxIntegral, root

    !-------------------------------------------------------------------
    ! Start:

    numSegments = size(x)
    if (numSegments <= 0) then
      numbers = 0.0
  return
    end if
    if (numSegments == 1) then
      numbers = x(1)
  return
    end if
    if (any(y < 0.0) .or. any(x(:numSegments-1) >= x(2:))) then
      numbers = 0.0
  return
    end if

    allocate(yy(numSegments), integratedSequence(numSegments))
    allocate(curvature(numSegments-1))

!    if (firstCall) then
!      firstCall = .false.
!    end if

    ! Rescale y values:

    maxIntegral = sum(y) - (y(1) + y(numSegments)) / 2.0
    ! ie = sum(i=1,numSegments-1)([y{i}+y{i+1}]/2)

    yy = y / maxIntegral

    ! Now load the integral and curvature vectors:

    integratedSequence(1) = 0.0
    curvature(1) = (yy(2) - yy(1)) / (x(2) - x(1)) / 2.0
    do i = 2, numSegments - 1
      integratedSequence(i) = integratedSequence(i-1) + (yy(i - 1) + yy(i))/2.0
      curvature(i) = (yy(i + 1) - yy(i)) / (x(i + 1) - x(i)) / 2.0
    end do
    integratedSequence(numSegments) = 1.0

    ! Now assign the random numbers:

    number = size(numbers)
    allocate(harvest(number))
    call getRandomNumbers(harvest)
    do n = 1, number
      bin = getBinNumber(integratedSequence, harvest(n))

      if (curvature(bin) == 0.0) then
        numbers(n) = x(bin) + (harvest(n) - integratedSequence(bin)) / yy(bin)

      else
        root = sqrt(yy(bin) * yy(bin) + 4.0 * curvature(bin) * (harvest(n)&
          - integratedSequence(bin)))

        numbers(n) = x(bin) + (-yy(bin) + root) / 2.0 / curvature(bin)
      end if
    end do

  end subroutine getSegmentRandoms

  !---------------------------------------------------------------------
  function poissonRandom(average)
  !---------------------------------------------------------------------
  ! Returns a random value with a Poissonian distribution. The algorithm is
  ! not efficient for large values of 'average'.

    implicit none

    real(single), intent(in)  :: average
    integer                   :: poissonRandom

    real(single) :: trial

    !-------------------------------------------------------------------
    ! Start:

    poissonRandom = 0
    trial = exponentialRandom()
    do while (trial < average)
      poissonRandom = poissonRandom + 1
      trial = trial + exponentialRandom()
    end do

  end function poissonRandom

  !---------------------------------------------------------------------
  subroutine getPoissonRandomsVector(average, numbers)
  !---------------------------------------------------------------------
  ! Returns a vector of values which have a Poissonian distribution.

    implicit none

    real(single), intent(in)  :: average
    integer,      intent(out) :: numbers(:)

    integer :: n, number

    !-------------------------------------------------------------------
    ! Start:

    number = size(numbers)
    do n = 1, number
      numbers(n) = poissonRandom(average)
    end do

  end subroutine getPoissonRandomsVector

  !---------------------------------------------------------------------
  subroutine getPoissonRandomsArray(average, numbers)
  !---------------------------------------------------------------------
  ! Returns a vector of values which have a Poissonian distribution.

    implicit none

    real(single), intent(in)  :: average
    integer,      intent(out) :: numbers(:,:)

    integer :: xiSize, yiSize, xi, yi

    !-------------------------------------------------------------------
    ! Start:

    xiSize = size(numbers, 1)
    yiSize = size(numbers, 2)
    do xi = 1, xiSize
      do yi = 1, yiSize
        numbers(xi, yi) = poissonRandom(average)
      end do
    end do

  end subroutine getPoissonRandomsArray

  !---------------------------------------------------------------------
  function minus5on2Random()
  !---------------------------------------------------------------------

    implicit none

    real(single) :: minus5on2Random

    integer, parameter :: hsize = 1000
    real(single) :: harvest(hsize)
    integer :: i
    save i, harvest
    data i /1/

    !-------------------------------------------------------------------
    ! Start:

    if (i .eq. 1) then
      call getMinus5on2Randoms(harvest)
    end if
    minus5on2Random = harvest(i)
    i = i + 1
    if (i .gt. hsize) i = 1

  end function minus5on2Random

  !---------------------------------------------------------------------
  subroutine getMinus5on2Randoms(numbers)
  !---------------------------------------------------------------------
  ! Returns a vector of values which have a distribution P given by
  ! P(x) = x^(-5/2) for x > 1.0. To get eg luminosities that follow the
  ! same distribution, multiply the returned values by the minimum detectable
  ! luminosity. It is not possible to return a distribution of luminosities
  ! down to zero luminosity because the integral of such a function is
  ! without bound.

    implicit none

    real(single), intent(out) :: numbers(:)
    real(single), allocatable :: harvest(:)
    integer :: n
    integer :: number
    real(single) :: minusTwoThirds
    logical :: firstCall
    save minusTwoThirds, firstCall
    data firstCall /.true./

    !-------------------------------------------------------------------
    ! Start:

    if (firstCall) then
      firstCall = .false.
      minusTwoThirds = -2.0 / 3.0
    end if

    number = size(numbers)
    allocate(harvest(number))

    call getRandomNumbers(harvest)
    do n = 1, number
      numbers(n) = (1.0 - harvest(n))**minusTwoThirds ! the subtraction of the
      ! random number from 1.0 is to prevent the occurrence of division by zero.
    end do

  end subroutine getMinus5on2Randoms

  !---------------------------------------------------------------------
  function powerRandom(power)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: power

    real(single) :: powerRandom
    integer, parameter :: hsize = 1000
    real(single) :: harvest(hsize)
    integer :: i
    save i, harvest
    data i /1/

    !-------------------------------------------------------------------
    ! Start:

    if (i .eq. 1) then
      call getPowerRandoms(power, harvest)
    end if
    powerRandom = harvest(i)
    i = i + 1
    if (i .gt. hsize) i = 1

  end function powerRandom

  !---------------------------------------------------------------------
  subroutine getPowerRandoms(power, numbers)
  !---------------------------------------------------------------------
  ! Returns a vector of values which have a distribution P given by
  ! P(x) = x^power. If power < -1, the distribution is given for for x > 1.0;
  ! if power >= 0, the distribution is given for 0 < x <= 1.0.

    implicit none

    real(single), intent(in) :: power
    real(single), intent(out) :: numbers(:)

    real(single), allocatable :: harvest(:)
    integer :: n
    integer :: number
    real(single) :: oneOverPowerPlusOne
    logical :: firstCall
    save oneOverPowerPlusOne, firstCall
    data firstCall /.true./

    !-------------------------------------------------------------------
    ! Start:

    if (firstCall) then
      firstCall = .false.
      oneOverPowerPlusOne = 1.0 / (power + 1.0)
    end if

    number = size(numbers)
    allocate(harvest(number))

    call getRandomNumbers(harvest)
    if (power < -1.0) then
      do n = 1, number
        numbers(n) = (1.0 - harvest(n))**oneOverPowerPlusOne ! the subtraction
        ! of the random number from 1.0 is to prevent the occurrence of
        ! division by zero. 1-rnd may equal 1 but cannot equal 0, hence
        ! the taking of negative powers of zero (ie, dividing by zero)
        ! is avoided.
      end do
    else if (power > -1.0) then
      do n = 1, number
        numbers(n) = harvest(n)**oneOverPowerPlusOne
      end do
    else ! power = -1, don't know how best to deal with this.
      numbers = -1.0
    end if

  end subroutine getPowerRandoms

  !---------------------------------------------------------------------
  subroutine getLogNLogSRandoms(x, yInt, hiPower, xSamples, origRandomsReturn&
  , status)
  !---------------------------------------------------------------------
  ! Note that it was decided not to implement a 'logNlogSrandom'
  ! function for this subroutine because of the need for a status return.
  ! There are too many things which can go wrong with this routine to
  ! allow a foolproof single-value return.
  !.....................................................................
  !
  ! The subroutine returns a vector of random values xSamples() which
  ! have a distribution P which, when integrated, follows a piecewise
  ! power law. The ith segment of the integrated piecewise function is
  ! defined between x(i) and x(i+1) and is of the form
  !
  !	log(yInt) = log(yInt(i)) + slope * (log(x) - log(x(i)))
  !
  ! where
  !	         log(yInt(i+1)) - log(yInt(i))
  !	slope = -------------------------------.
  !	            log(x(i+1)) - log(x(i))
  !
  ! Above x(N) the integrated function is defined by
  !
  !	log(yInt) = log(yInt(N)) + hiPower * (log(x) - log(x(N))).
  !
  ! Notes: (i) The x values must be in increasing order, and they cannot
  ! be negative. (ii) The y values must all be >0, and must never increase
  ! with increasing i. (iii) N must be >0. (iv) hiPower must be <0.
  !
  ! In the event of an error, the subroutine returns values of -1 for all
  ! elements of numbers().
  !.....................................................................

    implicit none

    real(single), intent(in)            :: x(:), yInt(size(x)), hiPower
    real(single), intent(out)           :: xSamples(:)
    real(single), intent(out), optional :: origRandomsReturn(size(xSamples))
    integer,      intent(out), optional :: status

    integer ::&
      localStatus, foreignStatus, numBins, numXsamples, i, bin
    real(single), allocatable ::&
      origRandoms(:), harvest(:), reciprocalSlopes(:), negYint(:)

logical :: testFlag=.false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, '  aaa getLogNLogSRandoms'

    localStatus = 0
    xSamples = -1

    numBins = size(x)
!print *, lbound(x), ubound(x)
    numXsamples = size(xSamples)

    allocate(origRandoms(numXsamples), harvest(numXsamples))
    allocate(reciprocalSlopes(numBins), negYint(numBins))

if (testFlag) print *, '  bbb getLogNLogSRandoms'
    call doLogNLogSChecks(x, yInt, hiPower, foreignStatus)
if (testFlag) print *, '  ccc getLogNLogSRandoms'

    if (foreignStatus < 0) then
      if (present(status)) status = foreignStatus
  return
    end if

    ! Calculate reciprocal slopes:

    do bin = 1, numBins - 1
      reciprocalSlopes(bin) = (log(x(   bin+1)) - log(x(   bin)))&
                            / (log(yInt(bin+1)) - log(yInt(bin)))
    end do
    reciprocalSlopes(numBins) = 1.0 / hiPower

    ! Main part of the routine:

if (testFlag) print *, '  ddd getLogNLogSRandoms'
    call getRandomNumbers(origRandoms)
    harvest = origRandoms * yInt(1)

    negYint = -yInt

    if (numBins>1) then
!******** better to adjust getBinNumber so that it behaves sensibly when there is only 1 bin edge.
      do i = 1, numXsamples
        bin = getBinNumber(negYint, -harvest(i)) ! in array_utils
!if (bin<=0) then
!print *, i, minval(negYint), maxval(negYint), -harvest(i)
!end if
        xSamples(i) = x(bin) * (harvest(i) / yInt(bin))**(reciprocalSlopes(bin))
      end do
    else
      do i = 1, numXsamples
        xSamples(i) = x(1) * (harvest(i) / yInt(1))**(reciprocalSlopes(1))
      end do
    end if

    if (present(origRandomsReturn)) origRandomsReturn = origRandoms
    if (present(status)) status = localStatus

if (testFlag) print *, '  zzz getLogNLogSRandoms'

  end subroutine getLogNLogSRandoms

  !---------------------------------------------------------------------
  subroutine getLogNLogSRandoms_20060509(x, yInt, hiPower, xSamples, origRandomsReturn&
  , status)
  !---------------------------------------------------------------------
  ! Note that it was decided not to implement a 'logNlogSrandom'
  ! function for this subroutine because of the need for a status return.
  ! There are too many things which can go wrong with this routine to
  ! allow a foolproof single-value return.
  !.....................................................................
  !
  ! The subroutine returns a vector of random values xSamples() which
  ! have a distribution P which, when integrated, follows a piecewise
  ! power law. The ith segment of the integrated piecewise function is
  ! defined between x(i) and x(i+1) and is of the form
  !
  !	log(yInt) = log(yInt(i)) + power * (log(x) - log(x(i)))
  !
  ! where
  !	         log(yInt(i+1)) - log(yInt(i))
  !	power = -------------------------------.
  !	            log(x(i+1)) - log(x(i))
  !
  ! Above x(N) the integrated function is defined by
  !
  !	log(yInt) = log(yInt(N)) + hiPower * (log(x) - log(x(N))).
  !
  ! Notes: (i) The x values must be in increasing order, and they cannot
  ! be negative. (ii) The y values must all be >0, and must never increase
  ! with increasing i. (iii) N must be >0. (iv) hiPower must be <0.
  !
  ! In the event of an error, the subroutine returns values of -1 for all
  ! elements of numbers().
  !.....................................................................

    implicit none

    real(single), intent(in)            :: x(:), yInt(size(x)), hiPower
    real(single), intent(out)           :: xSamples(:)
    real(single), intent(out), optional :: origRandomsReturn(size(xSamples))
    integer,      intent(out), optional :: status

    integer ::&
      localStatus, foreignStatus, numBins, numXsamples, i, bin
    real(single), allocatable ::&
      origRandoms(:), harvest(:), reciprocalSlopes(:), negYint(:)

logical :: testFlag=.true.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, '  aaa getLogNLogSRandoms'

    localStatus = 0
    xSamples = -1

    numBins = size(x)
print *, lbound(x), ubound(x)
    numXsamples = size(xSamples)

    allocate(origRandoms(numXsamples), harvest(numXsamples))
    allocate(reciprocalSlopes(numBins), negYint(numBins))

if (testFlag) print *, '  bbb getLogNLogSRandoms'
    call doLogNLogSChecks(x, yInt, hiPower, foreignStatus)
if (testFlag) print *, '  ccc getLogNLogSRandoms'

    if (foreignStatus < 0) then
      if (present(status)) status = foreignStatus
  return
    end if

    ! Calculate reciprocal slopes:

    do bin = 1, numBins - 1
      reciprocalSlopes(bin) = (log(x(   bin+1)) - log(x(   bin)))&
                            / (log(yInt(bin+1)) - log(yInt(bin)))
    end do
    reciprocalSlopes(numBins) = 1.0 / hiPower

    ! Main part of the routine:

if (testFlag) print *, '  ddd getLogNLogSRandoms'
    call getRandomNumbers(origRandoms)
    harvest = origRandoms * yInt(1)

    negYint = -yInt

print *, numBins
    if (numBins>1) then
      do i = 1, numXsamples
        bin = getBinNumber(negYint, -harvest(i)) ! in array_utils
if (bin<=0) then
print *, i, minval(negYint), maxval(negYint), -harvest(i)
end if
        xSamples(i) = x(bin) * (harvest(i) / yInt(bin))**(reciprocalSlopes(bin))
      end do
    else
      do i = 1, numXsamples
        xSamples(i) = x(1) * (harvest(i) / yInt(1))**(reciprocalSlopes(1))
      end do
    end if

    if (present(origRandomsReturn)) origRandomsReturn = origRandoms
    if (present(status)) status = localStatus

if (testFlag) print *, '  zzz getLogNLogSRandoms'

  end subroutine getLogNLogSRandoms_20060509

  !---------------------------------------------------------------------
  subroutine doLogNLogSChecks(x, yInt, hiPower, status)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)  :: x(:), yInt(size(x)), hiPower
    integer,      intent(out) :: status

    integer :: numPoints

    !-------------------------------------------------------------------
    ! Start:

    status = 0
    numPoints = size(x)

    if (numPoints < 1) then
      status = -1
  return
    end if

    if (numPoints > 1) then ! Check order of x values:
      if (any(x(2:) <= x(:numPoints-1))) then
        status = -2
  return
      end if

      if (any(yInt(2:) > yInt(:numPoints-1))) then
        status = -3
  return
      end if
    end if

    ! Check all x values >= 0:
    if (x(1) < 0.0) then
      status = -4
  return
    end if

    ! Check all yInt values > 0:
    if (any(yInt <= 0.0)) then
      status = -5
  return
    end if

!    if (hiPower >= -1.0) then
if (hiPower >= 0.0) then
      status = -6
  return
    end if

  end subroutine doLogNLogSChecks

  !---------------------------------------------------------------------
  subroutine getPowerSegmentRandoms(x, y, powers, numbers&
  , loPower, hiPower, status)
  !---------------------------------------------------------------------
  ! Note that it was decided not to implement a 'powerSegmentRandom'
  ! function for this subroutine because of the need for a status return.
  ! There are too many things which can go wrong with this routine to
  ! allow a foolproof single-value return.
  !.....................................................................
  !
  ! The subroutine returns a vector of random values which have a
  ! distribution P which follows a piecewise power law. The ith segment
  ! of this piecewise function is defined between x(i) and x(i+1) and is
  ! of the form
  !
  !	          y(i)
  !	y = ----------------- x**powers(i).
  !	     x(i)**powers(i)
  !
  ! Notes: (i) The amplitude is tied to the value at the lower x limit.
  ! (ii) There is nothing constraining the pieces of curve to join at the
  ! x boundaries. (iii) The value y(N) (where N=size(x)) is unused unless
  ! hiPower is supplied (see below). (iv) The x values must be in
  ! increasing order, and they cannot be negative - an error status is
  ! returned otherwise. (v) The y values all must be >=0, and at least
  ! 1 must be >0. Error conditions arise otherwise. (vi) Either N > 1
  ! with neither loPower or hiPower, or loPower and/or hiPower are supplied,
  ! in which case N may be >= 1. 
  !
  ! If the value loPower is supplied, P is assumed to continue below x=x(1)
  ! until x=0, with a functional form given by
  !
  !	         y(1)
  !	y = --------------- x**loPower.					(1)
  !	     x(1)**loPower
  !
  ! Notes: (i) This lower continuation is tied to the value on its
  ! _upper_ x boundary, in contrast to all the rest. (ii) loPower must
  ! be > -1. An error status is returned otherwise. (iii) If loPower is
  ! supplied, x(1) must be > 0. If loPower is not supplied, either (a)
  ! x(1) > 0, or (b) x(1) = 0 and power(i) > -1. An error status is
  ! returned if neither is the case.
  !
  ! If hiPower is supplied, P is assumed to continue on above x(N) until
  ! infinity, with a functional form
  !
  !	         y(N)
  !	y = ----------------x**hiPower.
  !	     x(N)**hiPower
  !
  ! Note: hiPower < -1.
  !
  ! In the event of an error, the subroutine returns values of -1 for all
  ! elements of numbers().
  !.....................................................................
  !
  ! This routine works as follows. We want to form a function
  !
  !	R = J(x)/J(inf),						(2)
  !
  ! where J is basically the stepwise integral of P, viz:
  !
  !	                       x
  !	                      /
  !	J(x) = J(x(i)) + A(i) | dx x**powers(i)				(3)
  !	                      /
  !	                     x(i)
  ! where
  !	                        y(i)
  !	A(i) is usually = -----------------,
  !	                   x(i)**powers(i)
  !
  ! with an exceptional form for A_lo, as shown above in equation 1.
  ! Provided that the x, y and powers obey the strictures given earlier,
  ! J(inf) exists and thus R(0) = 0 and R(inf) = 1. To return random
  ! numbers with the distribution P, evenly-distributed randoms are
  ! chosen, each is equated with R, and equations (3) and (2) are then
  ! inverted to give the x values. The inverse equation is given by
  !
  !	    [                           J(inf)
  !	x = [{powers(i) + 1}{R - R(i)} -------- +
  !	    [                            A(i)
  !
  !
  !	                        ]    (       1       )
  !	+ x(i)**{powers(i) + 1})] ** (---------------).
  !	                        ]    ( powers(i) + 1 )
  !
  !.....................................................................

    implicit none

    real(single), intent(in)            :: x(:), y(size(x)), powers(size(x)-1)
    real(single), intent(out)           :: numbers(:)
    real(single), intent(in),  optional :: loPower, hiPower
    integer,      intent(out), optional :: status

    integer ::&
      localStatus, foreignStatus, numPoints, numNumbers, i, bin, binEdge,&
      numBins
    real(single) :: pPlus1, Alo, Ahi, Jinf
    real(single), allocatable :: J(:), R(:), harvest(:), A(:) 

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0
    numbers = -1

    numPoints = size(x)
    numNumbers = size(numbers)
    allocate(harvest(numNumbers), A(numPoints-1))

    call doDiffPowerSegmentChecks(x, y, powers&
      , logical(present(loPower), bool), logical(present(hiPower), bool)&
      , foreignStatus)

    if (foreignStatus < 0) then
      if (present(status)) status = foreignStatus
  return
    end if

    numBins = numPoints - 1

    if (present(loPower)) then
      if (loPower <= -1.0) then
        if (present(status)) status = -10
  return
      end if

      Alo = y(1) / (x(1)**loPower)
      numBins = numBins + 1
    end if

    if (present(hiPower)) then
      if (hiPower >= -1.0) then
        if (present(status)) status = -11
  return
      end if

      Ahi = y(numPoints) / (x(numPoints)**hiPower)
      numBins = numBins + 1
    end if

    ! Calc A values:
    do i = 1, numPoints - 1
      A(i) = y(i) / (x(i)**powers(i))
    end do

    ! Calc total integral Jinf (and check > 0):
    allocate(J(numBins + 1), R(numBins + 1))
    binEdge = 1
    J(binEdge) = 0.0
    if (present(loPower)) then
      binEdge = binEdge + 1
      pPlus1 = loPower + 1.0
      J(binEdge) = J(binEdge-1) + (Alo / pPlus1) * (x(1)**pPlus1)
    end if

    do i = 1, numPoints
      binEdge = binEdge + 1
      pPlus1 = powers(i) + 1.0
      J(binEdge) = J(binEdge-1) + (A(i) / pPlus1)&
        * ((x(i+1)**pPlus1) - (x(i)**pPlus1))
    end do

    if (present(hiPower)) then
      binEdge = binEdge + 1
      pPlus1 = hiPower + 1.0
      J(binEdge) = J(binEdge-1) - (AHi / pPlus1) * (x(i)**pPlus1)

      ! The negative sign here is fine, because pPlus1
      ! must be < 0, thus the whole term ends up positive.
    end if

    Jinf = J(binEdge)
    if (Jinf <= 0.0) then
      if (present(status)) status = -12
  return
    end if

    ! Calc R boundaries (and check (tho should be needless with checks
    ! already done, still I'm a belt and braces man) that they are ordered
    ! in increasing value):

    R = J / Jinf
    if (any(R(2:) - R(:numBins) <= 0.0)) then
      status = -13
  return
    end if

    !...................................................................
    ! Main part of the routine:

    call getRandomNumbers(harvest)

    if (present(loPower)) then
      if (present(hiPower)) then
        do i = 1, numNumbers
          bin = getBinNumber(R, harvest(i)) ! in array_utils

          if (bin == 1) then
            pPlus1 = loPower + 1.0
            numbers(i) = (pPlus1 * harvest(i) * Jinf / Alo)**(1.0 / pPlus1)

          else if (bin == numBins) then
            pPlus1 = hiPower + 1.0
            numbers(i) = (pPlus1 * (harvest(i) - R(bin)) * Jinf / Ahi&
              + x(bin-1)**pPlus1)**(1.0 / pPlus1)

          else
            pPlus1 = powers(bin-1) + 1.0
            numbers(i) = (pPlus1 * (harvest(i) - R(bin)) * Jinf / A(bin-1)&
              + x(bin-1)**pPlus1)**(1.0 / pPlus1)
          end if
        end do

      else ! hiPower not present
        do i = 1, numNumbers
          bin = getBinNumber(R, harvest(i)) ! in array_utils

          if (bin == 1) then
            pPlus1 = loPower + 1.0
            numbers(i) = (pPlus1 * harvest(i) * Jinf / Alo)**(1.0 / pPlus1)

          else
            pPlus1 = powers(bin-1) + 1.0
            numbers(i) = (pPlus1 * (harvest(i) - R(bin)) * Jinf / A(bin-1)&
              + x(bin-1)**pPlus1)**(1.0 / pPlus1)
          end if
        end do
      end if

    else ! loPower not present
      if (present(hiPower)) then
        do i = 1, numNumbers
          bin = getBinNumber(R, harvest(i)) ! in array_utils

          if (bin == numBins) then
            pPlus1 = hiPower + 1.0
            numbers(i) = (pPlus1 * (harvest(i) - R(bin)) * Jinf / Ahi&
              + x(bin)**pPlus1)**(1.0 / pPlus1)

          else
            pPlus1 = powers(bin) + 1.0
            numbers(i) = (pPlus1 * (harvest(i) - R(bin)) * Jinf / A(bin)&
              + x(bin)**pPlus1)**(1.0 / pPlus1)
          end if
        end do

      else ! hiPower not present
        do i = 1, numNumbers
          bin = getBinNumber(R, harvest(i)) ! in array_utils

          pPlus1 = powers(bin) + 1.0
          numbers(i) = (pPlus1 * (harvest(i) - R(bin)) * Jinf / A(bin)&
            + x(bin)**pPlus1)**(1.0 / pPlus1)
        end do
      end if
    end if

    if (present(status)) status = localStatus

  end subroutine getPowerSegmentRandoms

  !---------------------------------------------------------------------
  subroutine doDiffPowerSegmentChecks(x, y, powers&
  , presentLoPower, presentHiPower, status)
  !---------------------------------------------------------------------
!*** bit of a waste of space here because int and diff differ only in 2 characters.

    implicit none

    real(single), intent(in)  :: x(:), y(size(x)), powers(size(x)-1)
    logical(bool),intent(in)  :: presentLoPower, presentHiPower
    integer,      intent(out) :: status

    integer :: numPoints

    !-------------------------------------------------------------------
    ! Start:

    numPoints = size(x)

    if (presentLoPower .or. presentHiPower) then
      if (numPoints < 1) then
        status = -1
  return
      end if

      if (numPoints > 1) then ! Check order of x values:
        if (any(x(2:) <= x(:numPoints-1))) then
          status = -2
  return
        end if
      end if

    else
      if (numPoints < 2) then
        status = -3
  return
      end if

      ! Check order of x values:
      if (any(x(2:) <= x(:numPoints-1))) then
        status = -4
  return
      end if
    end if

    ! Check all x values >= 0:
    if (x(1) < 0.0) then
      status = -5
  return
    end if

    if (presentLoPower) then
      ! Check all x values > 0:
      if (x(1) <= 0.0) then
        status = -6
  return
      end if

    else
      if (powers(1) <= -1.0 .and. x(1) == 0.0) then
        status = -7
  return
      end if
    end if

    ! Check all y values >= 0:
    if (any(y < 0.0)) then
      status = -8
  return
    end if

    ! Check at least 1 y value > 0:
    if (all(y <= 0.0)) then
      status = -9
  return
    end if

  end subroutine doDiffPowerSegmentChecks

  !---------------------------------------------------------------------
  subroutine doIntPowerSegmentChecks(x, y, powers&
  , presentLoPower, presentHiPower, status)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)  :: x(:), y(size(x)), powers(size(x)-1)
    logical(bool),intent(in)  :: presentLoPower, presentHiPower
    integer,      intent(out) :: status

    integer :: numPoints

    !-------------------------------------------------------------------
    ! Start:

    numPoints = size(x)

    if (presentLoPower .or. presentHiPower) then
      if (numPoints < 1) then
        status = -1
  return
      end if

      if (numPoints > 1) then ! Check order of x values:
        if (any(x(2:) <= x(:numPoints-1))) then
          status = -2
  return
        end if
      end if

    else
      if (numPoints < 2) then
        status = -3
  return
      end if

      ! Check order of x values:
      if (any(x(2:) <= x(:numPoints-1))) then
        status = -4
  return
      end if
    end if

    ! Check all x values >= 0:
    if (x(1) < 0.0) then
      status = -5
  return
    end if

    if (presentLoPower) then
      ! Check all x values > 0:
      if (x(1) <= 0.0) then
        status = -6
  return
      end if

    else
      if (powers(1) <= 0.0 .and. x(1) == 0.0) then
        status = -7
  return
      end if
    end if

    ! Check all y values >= 0:
    if (any(y < 0.0)) then
      status = -8
  return
    end if

    ! Check at least 1 y value > 0:
    if (all(y <= 0.0)) then
      status = -9
  return
    end if

  end subroutine doIntPowerSegmentChecks

  !---------------------------------------------------------------------
  function sineRandom(minX, maxX)
  !---------------------------------------------------------------------
!***** put into ssclib/random_aux! IMS 2006-05-16

    implicit none

    real(single), intent(in), optional :: minX, maxX
    real(single)                       :: sineRandom

    integer, parameter :: hsize = 1000
    real(single) :: harvest(hsize)
    integer :: i
    save i, harvest
    data i /1/

    !-------------------------------------------------------------------
    ! Start:

    if (i == 1) then
      call getSineRandoms(harvest, minX, maxX)
    end if
    sineRandom = harvest(i)
    i = i + 1
    if (i > hsize) i = 1

  end function sineRandom

  !---------------------------------------------------------------------
  subroutine getSineRandoms(numbers, minX, maxX, status)
  !---------------------------------------------------------------------
!***** replace the present one in ssclib/random_aux with this! IMS 2006-05-16

  ! Returns a vector of values which have a distribution P given by
  ! P(x) = sine(x) for 0 <= minX <= x < maxX < Pi. This can be used
  ! to get the radius or codeclination values of points which have a smooth
  ! distribution over a conical segment of a sphere of cone angle maxX radians.

!*** Modified it to avoid trouble at angles near zero, where the random numbers in harvest were being squashed into a narrow range just below 1.0 and thus losing precision. Should modify it to use sqrt in a small interval close to zero. IMS 20060809.

    implicit none

    real(single), intent(out)           :: numbers(:)
    real(single), intent(in),  optional :: minX, maxX
    integer,      intent(out), optional :: status

    real(single), allocatable :: harvest(:)
    integer :: numPoints
!, i
    real(single) :: x0, x1, c0, c1
!, minn
real(double), allocatable :: args(:)

    !-------------------------------------------------------------------
    ! Start:

    if (present(status)) status = 0
    numbers = 0.0 ! default

    x0 = decideOption(minX, 0.0)
    x1 = decideOption(maxX, real(Pi,single)) ! in ssc_misc.

    if (x1 > Pi) then
      if (present(status)) status = 1
      x1 = Pi
    end if
    if (x0 < 0.0) then
      if (present(status)) status = 2
      x0 = 0.0
    end if

    c0 = 1.0-cos(x0)
    c1 = 1.0-cos(x1)
!    if (c1<-1) c1 = nearest(-1.0,1.0)
    if (c0 >= c1) then
      if (present(status)) status = -1
  return
    end if

    numPoints = size(numbers)
    allocate(harvest(numPoints))
allocate(args(numPoints))
    call getRandomNumbers(harvest)
args = fracVal(c0, c1, harvest)
!    numbers = acos(fracVal(c1, c0, harvest)) ! in ssc_misc.
numbers = acos(1.0d0-args) ! in ssc_misc.

!minn = Pi
!do i = 1, numPoints
!if (numbers(i)==0.0) cycle
!  if (numbers(i) < minn) minn = numbers(i)
!  if (numbers(i) < 8.0e-4) print *, i, harvest(i), args(i), numbers(i)
!end do
!print *, minn, x0, x1, c0, c1

  end subroutine getSineRandoms

  !---------------------------------------------------------------------
  subroutine getSineRandoms_20060809(numbers, minX, maxX, status)
  !---------------------------------------------------------------------
!***** replace the present one in ssclib/random_aux with this! IMS 2006-05-16

  ! Returns a vector of values which have a distribution P given by
  ! P(x) = sine(x) for 0 <= minX <= x < maxX < Pi. This can be used
  ! to get the radius or codeclination values of points which have a smooth
  ! distribution over a conical segment of a sphere of cone angle maxX radians.

    implicit none

    real(single), intent(out)           :: numbers(:)
    real(single), intent(in),  optional :: minX, maxX
    integer,      intent(out), optional :: status

    real(single), allocatable :: harvest(:)
    integer :: numPoints, i
    real(single) :: x0, x1, c0, c1, minn
real(single), allocatable :: args(:)

    !-------------------------------------------------------------------
    ! Start:

    if (present(status)) status = 0
    numbers = 0.0 ! default

    x0 = decideOption(minX, 0.0)
    x1 = decideOption(maxX, real(Pi,single)) ! in ssc_misc.

    if (x1 > Pi) then
      if (present(status)) status = 1
      x1 = Pi
    end if
    if (x0 < 0.0) then
      if (present(status)) status = 2
      x0 = 0.0
    end if

    c0 = cos(x0)
    c1 = cos(x1)
    if (c1<-1) c1 = nearest(-1.0,1.0)
    if (c0 <= c1) then
      if (present(status)) status = -1
  return
    end if

    numPoints = size(numbers)
    allocate(harvest(numPoints))
allocate(args(numPoints))
    call getRandomNumbers(harvest)
args = fracVal(c1, c0, harvest)
!    numbers = acos(fracVal(c1, c0, harvest)) ! in ssc_misc.
numbers = acos(args) ! in ssc_misc.

minn = Pi
do i = 1, numPoints
if (numbers(i)==0.0) cycle
  if (numbers(i) < minn) minn = numbers(i)
  if (numbers(i) < 8.0e-4) print *, i, harvest(i), args(i), numbers(i)
end do
print *, minn, x0, x1, c0, c1

  end subroutine getSineRandoms_20060809

  !---------------------------------------------------------------------
  subroutine getSineRandoms_20060516(maxX, numbers, status)
  !---------------------------------------------------------------------
  ! Returns a vector of values which have a distribution P given by
  ! P(x) = sine(x) / (1 - cos(maxX)) for 0.0 < x < maxX. This can be used
  ! to get the radius or codeclination values of points which have a smooth
  ! distribution over a conical segment of a sphere of cone angle maxX radians.

    implicit none

    real(single), intent(in)            :: maxX
    real(single), intent(out)           :: numbers(:)
    integer,      intent(out), optional :: status

    real(single), allocatable :: harvest(:)
    integer :: n
    integer :: number
    real(single) :: k
    logical :: firstCall
    save k, firstCall
    data firstCall /.true./

    !-------------------------------------------------------------------
    ! Start:

    if (present(status)) status = 0
    numbers = 0.0 ! default
    if (maxX .gt. PiOverTwo) then
      if (present(status)) status = -1
  return
    end if
    if (maxX .le. 0.0) then
      if (present(status)) status = -2
  return
    end if

    if (firstCall) then
      firstCall = .false.
      k = calcK(maxX)
    end if

    number = size(numbers)
    allocate(harvest(number))

    call getRandomNumbers(harvest)
    do n = 1, number
      numbers(n) = asin(sqrt(k * harvest(n) * (2.0 - k * harvest(n))))
    end do

  end subroutine getSineRandoms_20060516

  !---------------------------------------------------------------------
  function calcK(maxX) result(k)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: maxX
    real(single)             :: k

    real(single) :: myPrecision, sinXonTwo, kOld, kNew

    !-------------------------------------------------------------------
    ! Start:

    myPrecision = 50.0 ** (-precision(maxX))

    if (maxX .le. 0.0) then
      k = 0.0
  return
    end if

    if (maxX .lt. 0.1) then ! use an iterative expression to get k.
    ! The 0.1 is essentially arbitrary.
      sinXonTwo = sin(maxX) * sin(maxX) / 2.0
      kOld = sinXonTwo
      do
        kNew = sinXonTwo + kOld * kOld
      if ((kNew - kOld) / kOld .le. myPrecision) exit
        kOld = kNew
      end do
      k = kOld

    else
      k = 1.0 - sqrt(1.0 - sin(maxX) * sin(maxX))
    end if

  end function calcK

  !---------------------------------------------------------------------
  function exponentialRandom()
  !---------------------------------------------------------------------

    implicit none

    real(single) :: exponentialRandom
    integer, parameter :: hsize = 1000
    real(single) :: harvest(hsize)
    integer :: i
    save i, harvest
    data i /1/

    !-------------------------------------------------------------------
    ! Start:

    if (i .eq. 1) then
      call getExponentialRandoms(harvest)
    end if
    exponentialRandom = harvest(i)
    i = i + 1
    if (i .gt. hsize) i = 1

  end function exponentialRandom

  !---------------------------------------------------------------------
  subroutine getExponentialRandoms(numbers)
  !---------------------------------------------------------------------
  ! Returns a vector of values which have an exponential probability
  ! distribution with a mean value of 1.0.

    implicit none

    real(single), intent(out) :: numbers(:)

    real(single), allocatable :: harvest(:)
    integer :: n
    integer :: number

    !-------------------------------------------------------------------
    ! Start:

    number = size(numbers)
    allocate(harvest(number))

    call getRandomNumbers(harvest)
    do n = 1, number
      numbers(n) = -log(1.0 - harvest(n)) ! the subtraction of the
      ! random number from 1.0 is to prevent the occurrence of zero-
      ! valued argument to log().
    end do

  end subroutine getExponentialRandoms

  !---------------------------------------------------------------------
  function gaussianRandom()
  !---------------------------------------------------------------------

    implicit none

    real(single) :: gaussianRandom
    integer, parameter :: hsize = 1000
    real(single) :: harvest(hsize)
    integer :: i
    save i, harvest
    data i /1/

    !-------------------------------------------------------------------
    ! Start:

    if (i .eq. 1) then
      call getGaussianRandoms(harvest)
    end if
    gaussianRandom = harvest(i)
    i = i + 1
    if (i .gt. hsize) i = 1

  end function gaussianRandom

  !---------------------------------------------------------------------
  subroutine getGaussianRandoms(numbers)
  !---------------------------------------------------------------------
  ! Returns a vector of values which have an gaussian probability
  ! distribution about zero with standard deviation equal to 1.0.

    implicit none

    real(single), intent(out) :: numbers(:)

    integer :: n, number
    real(single) fac, rsq, v1, v2

    !-------------------------------------------------------------------
    ! Start:

    number = size(numbers)

    if (number / 2.0 .eq. floor(number / 2.0)) then ! even
      do n = 1, number - 1, 2
        do
          v1 = 2.0 * randomNumber() - 1.0
          v2 = 2.0 * randomNumber() - 1.0
          rsq = v1 * v1 + v2 * v2

        if (rsq .gt. 0.0 .and. rsq .lt. 1.0) exit
        end do

        fac = sqrt(-2.0 * log(rsq) / rsq)
        numbers(n)     = v1 * fac
        numbers(n + 1) = v2 * fac
      end do

    else ! odd
      do n = 1, number - 2, 2
        do
          v1 = 2.0 * randomNumber() - 1.0
          v2 = 2.0 * randomNumber() - 1.0
          rsq = v1 * v1 + v2 * v2

        if (rsq .gt. 0.0 .and. rsq .lt. 1.0) exit
        end do

        fac = sqrt(-2.0 * log(rsq) / rsq)
        numbers(n)     = v1 * fac
        numbers(n + 1) = v2 * fac
      end do

      do
        v1 = 2.0 * randomNumber() - 1.0
        v2 = 2.0 * randomNumber() - 1.0
        rsq = v1 * v1 + v2 * v2

      if (rsq .gt. 0.0 .and. rsq .lt. 1.0) exit
      end do

      fac = sqrt(-2.0 * log(rsq) / rsq)
      numbers(number) = v1 * fac
    end if

  end subroutine getGaussianRandoms

  !---------------------------------------------------------------------
  function randoms(number)
  !---------------------------------------------------------------------

    implicit none

    integer, intent(in) :: number
    real(single) :: randoms(number)

    !-------------------------------------------------------------------
    ! Start

    call getRandomNumbers(randoms)

  end function randoms

  !---------------------------------------------------------------------
  function makeRandomSequence(duration, rate) result(randomSequence)
  !---------------------------------------------------------------------
  ! This function returns a random sequence of 'times' which occur at an
  ! average rate and for a duration given in the function arguments.
  !
  ! The difficulty in implementing this function lies in the fact that
  ! the final size of the returned sequence is not known until the last
  ! time is calculated (at least, with the present algorithm). So the amount
  ! of memory needed cannot be allocated a priori. This is partially
  ! overcome by constructing an array of pointers (called here 'chunk')
  ! and allocating memory to each component of this array as it becomes
  ! necessary. The values are then copied in sequence from the allocated
  ! components into the returned pointer.

    implicit none

    real(single), intent(in) :: duration, rate
    real(double), pointer :: randomSequence(:)
!**** why return double for single? This is just asking for trouble.

    type :: bufferType
      real(double), pointer :: buffer(:)
    end type
    integer, parameter :: maxNumChunks = 20
    type(bufferType) :: chunk(maxNumChunks)
    integer :: bufferSize, chunkNum, n, lastFullChunk, start, finis
    real(double) :: time, timeIncrement

    !-------------------------------------------------------------------
    ! Start

    bufferSize = min(1048576, max(1024, nint(duration * rate)))

    chunkNum = 1
    allocate(chunk(chunkNum)%buffer(bufferSize))
    n = 0
    time = 0.0d0
    do
      timeIncrement = exponentialRandom() / rate
      time = time + timeIncrement
    if (time > duration) exit

      n = n + 1
      if (n > bufferSize) then
        chunkNum = chunkNum + 1
        if (chunkNum > maxNumChunks) then
          n = 0
    exit
        end if

        allocate(chunk(chunkNum)%buffer(bufferSize))
        n = 1
      end if

      chunk(chunkNum)%buffer(n) = time
    end do

    lastFullChunk = chunkNum - 1

    allocate(randomSequence(lastFullChunk * bufferSize + n))
    start = 1
    finis = bufferSize
    do chunkNum = 1, lastFullChunk
      randomSequence(start:finis) = chunk(chunkNum)%buffer
      start = start + bufferSize
      finis = finis + bufferSize
      deallocate(chunk(chunkNum)%buffer)
    end do
    if (n > 0) then
      randomSequence(start:start + n - 1) = chunk(lastFullChunk + 1)%buffer(:n)
      deallocate(chunk(lastFullChunk + 1)%buffer)
    end if

  end function makeRandomSequence

  !---------------------------------------------------------------------
  logical(bool) function coinIsHeads()
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

    if (randomNumber() > 0.5) then
      coinIsHeads = .true.
    else
      coinIsHeads = .false.
    end if

  end function coinIsHeads

  !---------------------------------------------------------------------
  subroutine getRandomOrder(indices)
  !---------------------------------------------------------------------
  ! The returned vector indices contains integers from 1 to size(indices),
  ! but their order is random.

    implicit none

    integer, intent(out) :: indices(:)

    real(single), allocatable :: numbers(:)
    integer :: numIndices

    !-------------------------------------------------------------------
    ! Start

    numIndices = size(indices)
    allocate(numbers(numIndices))
    call getRandomNumbers(numbers)
    call quickSort(numbers, indices)

  end subroutine getRandomOrder

  !---------------------------------------------------------------------
  function randomNumber()
  !---------------------------------------------------------------------

    implicit none

    real(single) :: randomNumber
    integer, parameter :: hsize = 1000
    real(single) :: harvest(hsize)
    integer :: i
    save i, harvest
    data i /1/

    !-------------------------------------------------------------------
    ! Start:

    if (i .eq. 1) then
      call getRandomNumbers(harvest)
    end if
    randomNumber = harvest(i)
    i = i + 1
    if (i .gt. hsize) i = 1

  end function randomNumber

  !---------------------------------------------------------------------
  subroutine getRandomNumbers(numbers)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(out) :: numbers(:)

    integer :: numRands, randNum

    !-------------------------------------------------------------------
    ! Start:

    call SAS_randSetMinMaxRange(0.0d0, 1.0d0, CLOSED_OPEN)
    numRands = size(numbers)
    do randNum = 1, numRands
      numbers(randNum) = SAS_rand() ! in utils/random.f90.
    end do

  end subroutine getRandomNumbers

  !---------------------------------------------------------------------
  subroutine getRandomNumbers_20031111(numbers)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(out) :: numbers(:)

    !-------------------------------------------------------------------
    ! Start:

   call random_number(numbers) ! f90 intrinsic routine. Probably not very good.

!    call SAS_rand(numbers) ! in utils. Not yet available in rel track (IMS 2002-12-05).

  end subroutine getRandomNumbers_20031111

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module l_random_aux
