!+
! Module: srcmap_mod
!
! Authors:  Ian Stewart - JBO
!          
! Description: Makes an image from a source list.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module srcmap_mod

  use dal
  use param  
  use errorHandling
  use dataFilter	! in selectlib

  use array_utils	! for addPatch.
  use constants
  use coordinate
  use dal_aux
!use l_dal_aux_2
use dump
  use save_image
  use ssc_misc
  use wcs_aux
!use temp_ssclib

  implicit none 

  private 

  ! Command-line parameters:
  character(1000) ::&
    selectionExpression=''
  character(256) ::&
    srcListSetTabName='',&
    maskSetName='',&
    templateImageSetName='',&
    outImageSetName='',&
    tempImageSetName='',&
    tempSetName=''
  character(20) ::&
    raColName='',&
    decColName='',&
    rateColName='',&
    gaussApproxStyle='',&
    psfStyle='',&
    widthColName=''
  logical ::&
!    withCutoff=.false.,&
    withGaussianApprox=.false.,&
    withMaskSet=.false.,&
    normalize=.false.,&
    vignet=.false.,&
    withFluxOffset=.false.
  real(single) ::&
!    ecf=0.0d0,&
!    normalizedFlux=0.0d0,&
!    psfEnergyFraction=0.0d0,&
!    cutoffFlux=0.0,&
    fluxScale=0.0,&
    gaussFrac=0.0,&
!    r0Pixels2=0.0,&
!    sigmaPixels=0.0,&
    r0Arcsec=0.0,&
    sigmaArcsec=0.0,&
    alpha=0.0,&
    fluxOffset=0.0
  integer ::&
    numSubPixelDivs = 0,&
    halfISize=0

  real(single) ::&
    r0Pixels2=0.0,&
    sigmaPixels=0.0

!  real(single), parameter :: patchWidthInSigmas = 7.0
  integer, parameter :: patchHalfWidthPixels = 7

  public srcmap

contains 

  !---------------------------------------------------------------------
  subroutine srcmap
  !---------------------------------------------------------------------
  ! This makes a simulated sky map from the source list, convolving each
  ! source position with a Point Spread Function (PSF).
  !
  ! The sky map is created in the projection plane implicit in the wcs
  ! keywords of the mask image. Sky map pixels are left set to zero
  ! where the mask is zero.
  !
  ! If --withfluxcut is set, only sources which have FLUX values less
  ! than --fluxcut contribute to the image.
  !.....................................................................

    implicit none

    type(DataSetT) :: templateSet, srcListSet, imageSet
    type(ArrayT) :: templateArray, imageArray
    type(WcsT) :: wcs
    integer, pointer :: imageDims(:)
    integer(int16) :: xiSize, yiSize
    integer ::&
      numSources, patchXiStart, patchXiFinis, patchYiStart,&
      patchYiFinis, src, xi, yi, subXi, subYi, numSuccessful
    logical, pointer :: mask(:,:)
    character(256) :: srcListSetName
    character(20) :: srcListTabName
    type(TableT) :: sourceTab
    type(ColumnT) ::&
      srcRaDegCol, srcDecDegCol, fluxCol, widthCol
!    real(double), pointer :: srcRaDeg(:), srcDecDeg(:)
real(double), allocatable :: srcRaDeg(:), srcDecDeg(:), flux(:), widths(:)
    real(single), pointer ::&
!      flux(:), widths(:),&
      patchHyperCube(:,:,:,:), patch(:,:)
    real(single), allocatable ::&
      scaledFlux(:), srcXpixels(:), srcYpixels(:), image(:,:)
    real(double), allocatable :: srcTanX(:), srcTanY(:)
!    real(double) :: xPixelSizeWorld, yPixelSizeWorld
    real(single) :: x, y, pixelAreaArcsec
!, pixelXSizeArcsec
!, r0Pixels
logical :: testFlag = .false.
character(20) :: junkname

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, 'aaa'

    call readParameters

    !...................................................................
    ! Read image template and mask details:

    templateSet = dataSet(templateImageSetName, READ)
    templateArray = array(templateSet, 0)
    call readWcs(templateSet, templateArray, wcs) ! in ssclib/wcs_aux.

    imageDims => dimensions(templateArray)
    xiSize = imageDims(0)
    yiSize = imageDims(1)

    if (withMaskSet) then
      call readArrayData(maskSetName, mask) ! in ssclib/dal_aux.
      if (size(mask, 1) .ne. xiSize .or. size(mask, 2) .ne. yiSize)&
call error('sizeMismatch',"")
    else
      allocate(mask(xiSize,yiSize))
      mask = .true.
    end if

    call release(templateArray)
    call release(templateSet)

    ! Calculate r0 and sigma in units of image pixels:
    !
!!***** CAUTION! The following assumes (i) that image pixels are square, and (ii) that WCS world coords are degrees.
!    call convertNewToOld(wcs, xPixelSizeWorld, yPixelSizeWorld) ! in ssclib/wcs_aux.
!    pixelXSizeArcsec = abs(xPixelSizeWorld) * 3600.0 ! =4.0 for xmm product images.
!    r0Pixels = r0Arcsec / pixelXSizeArcsec
!    r0Pixels2 = r0Pixels * r0Pixels
!    sigmaPixels = sigmaArcsec / pixelXSizeArcsec
pixelAreaArcsec = wcsPixelAreaWorld(wcs) * 3600.0 * 3600.0 
!***** CAUTION! This assumes that WCS world coords are degrees.
r0Pixels2 = r0Arcsec * r0Arcsec / pixelAreaArcsec
sigmaPixels = sigmaArcsec / sqrt(pixelAreaArcsec)

    !...................................................................
    ! Open source set and load the columns:

if (testFlag) print *, 'bbb'
    call splitSetTabName(srcListSetTabName, srcListSetName, srcListTabName)
    ! in ssclib/dal_aux.
    srcListSet = clone(srcListSetName, tempSetName, TEMP)

    sourceTab = table(srcListSet, srcListTabName, MODIFY)
    call initFilter(selectionExpression)
    numSources = doFilter(sourceTab)
if (testFlag) print *, 'ccc'

    if (numSources <= 0)&
      call warning('noSourcesSelected'&
      , 'No sources passed the specified filtering expression.')

!    srcRaDegCol  = column(sourceTab, raColName)
!    srcDecDegCol = column(sourceTab, decColName)
!    fluxCol      = column(sourceTab, rateColName)
!
!    srcRaDeg  => real64Data(srcRaDegCol)
!    srcDecDeg => real64Data(srcDecDegCol)
!    flux      => real32Data(fluxCol) ! flux density at 1 keV.
allocate(srcRaDeg(numSources), srcDecDeg(numSources), flux(numSources))
call readColDataToFixed(sourceTab, raColName,   srcRaDeg)
call readColDataToFixed(sourceTab, decColName,  srcDecDeg)
call readColDataToFixed(sourceTab, rateColName, flux)

    allocate(scaledFlux(numSources))
    scaledFlux = flux * fluxScale

if (testFlag) print *, 'ddd'
    !...................................................................
    ! Convert source Ra/Dec to image coordinates:

    allocate(srcTanX(numSources), srcTanY(numSources))
    call raDecToTan(srcRaDeg*PiOn180, srcDecDeg*PiOn180, wcs&
      , srcTanX, srcTanY) ! in ssclib/coordinate

    allocate(srcXpixels(numSources), srcYpixels(numSources))
    call tanToPixels(wcs, srcTanX, srcTanY, srcXpixels, srcYpixels)
    ! in ssclib/wcs_aux
    deallocate(srcTanX, srcTanY)

!    call release(srcRaDegCol)
!    call release(srcDecDegCol)

    !...................................................................
    ! Now work through the list of sources and add a correctly-normalized
    ! PSF image for each one that falls on a non-zero pixel of the mask:

if (testFlag) print *, 'eee'
    allocate(image(xiSize, yiSize))
    image = 0.0

    select case(psfStyle)
    case('uniform') ! ie, each PSF is the same.
      if (withGaussianApprox) then
        select case(gaussApproxStyle)
        case('interpolate')
call error('notYetSupported',"")
        case('subpixels')
if (testFlag) print *, 'fff'
          allocate(patchHyperCube(-halfISize:halfISize,-halfISize:halfISize&
            , 0:numSubPixelDivs-1, 0:numSubPixelDivs-1))

          call calcCubeOfJitteredPsfPatches(patchHyperCube)

          patchXiStart = lbound(patchHyperCube, 1)
          patchXiFinis = ubound(patchHyperCube, 1)
          patchYiStart = lbound(patchHyperCube, 2)
          patchYiFinis = ubound(patchHyperCube, 2)

numSuccessful = 0
          do src = 1, numSources
            errStr = sprintEvery(src, numSources, fraction=0.05) ! in ssclib/ssc_misc.
            if (errStr .ne. '') then
              errStr = '     '//errStr
              call message(appMsg, verbose)
            end if

            x = srcXpixels(src) + 0.5
            y = srcYpixels(src) + 0.5
            xi = floor(x)
            yi = floor(y)

          if (xi < 1 .or. xi > xiSize .or. yi < 1 .or. yi > yiSize) cycle
          if (.not. mask(xi, yi)) cycle
numSuccessful = numSuccessful + 1
if (testFlag) print *, src, numSuccessful

            allocate(patch(patchXiStart+xi:patchXiFinis+xi&
             , patchYiStart+yi:patchYiFinis+yi))
            subXi = floor(numSubPixelDivs * (x - xi))
            subYi = floor(numSubPixelDivs * (y - yi))
            patch = patchHyperCube(:,:,subXi,subYi)

            call addPatch(image, patch, scaledFlux(src)) ! in ssclib/array_utils.
            deallocate(patch)
          end do ! src
!print *, 'numSuccessful, numSources:', numSuccessful, numSources

if (testFlag) print *, 'ggg'
          deallocate(patchHyperCube)

        case default
call error('badApproxStyle',"")
        end select

      else ! don't approximate the gaussians: calculate each one exactly. (Well, almost.)
        do src = 1, numSources
          errStr = sprintEvery(src, numSources, fraction=0.05) ! in ssclib/ssc_misc.
          if (errStr .ne. '') then
            errStr = '     '//errStr
            call message(appMsg, verbose)
          end if

          xi = nint(srcXpixels(src))
          yi = nint(srcYpixels(src))

        if (xi < 1 .or. xi > xiSize .or. yi < 1 .or. yi > yiSize) cycle
        if (.not. mask(xi, yi)) cycle

          allocate(patch(-halfISize+xi:halfISize+xi,-halfISize+yi:halfISize+yi))
          call getPsfPatch(srcXpixels(src), srcYpixels(src), patch)
!print *, src, scaledFlux(src)
!write (junkname, '(a,i2.2,a)') 'patch_', src, '.fits'
!call dumpImageToFits(patch, junkname)
          call addPatch(image, patch, scaledFlux(src)) ! in ssclib/array_utils.
          deallocate(patch)
        end do ! src
      end if ! withGaussianApprox

    ! psfStyle:
    case('column') ! => width of PSF is read from the source list.
!      widthCol = column(sourceTab, widthColName)
!      widths => real32Data(widthCol)
allocate(widths(numSources))
call readColDataToFixed(sourceTab, widthColName, widths)

      gaussFrac = 1.0
      r0Pixels2 = 1.0 ! should not be used, but give them
      alpha = 1.0     ! ok dummy values just in case...
      do src = 1, numSources
        errStr = sprintEvery(src, numSources, fraction=0.05) ! in ssclib/ssc_misc.
        if (errStr .ne. '') then
          errStr = '     '//errStr
          call message(appMsg, verbose)
        end if

        xi = nint(srcXpixels(src))
        yi = nint(srcYpixels(src))

      if (xi < 1 .or. xi > xiSize .or. yi < 1 .or. yi > yiSize) cycle
      if (.not. mask(xi, yi)) cycle

!        if (widths(src-1)<=0.0) then
        if (widths(src)<=0.0) then
          call warning('nonPositiveGaussianWidth',"")
      cycle
        end if

!        sigmaPixels = widths(src-1)
        sigmaPixels = widths(src)
        allocate(patch(-halfISize:halfISize,-halfISize:halfISize))
        call getPsfPatch(srcXpixels(src), srcYpixels(src), patch)

        call addPatch(image, patch, scaledFlux(src)) ! in ssclib/array_utils.
        deallocate(patch)
      end do ! src
!      call release(widthCol)

    case default
call error('badPsfStyle',"")
    end select
if (testFlag) print *, 'hhh'

!    call release(fluxCol)
    call release(sourceTab)
    deallocate(mask)

    if (withFluxOffset) then
      image = image + fluxOffset * pixelAreaArcsec
    end if

    !...................................................................
    ! Save the output:

if (testFlag) print *, 'iii'
!print *, 'maxval(image)', maxval(image)

    call saveImage(image, templateImageSetName, outImageSetName, tempImageSetName)
    ! in ssclib/save_image.
    ! Copies over all keywords (eg wcs, INSTRUM) in the templateImage.

    imageSet = dataSet(outImageSetName, MODIFY)
    imageArray = array(imageSet, 0)
    call setAttribute(imageArray, 'FLUXSCAL',  fluxScale, ''&
      , 'The source '//trim(rateColName)//' values were multiplied by this quantity.')

    if (hasBlock(srcListSet, 'SRCSPECS'))&
      call copyBlock(imageSet, block(srcListSet, 'SRCSPECS'))

    call release(imageSet)
    call release(srcListSet)

if (testFlag) print *, 'zzz'
  end subroutine srcmap

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

!    real(single) :: r0Pixels

    !-------------------------------------------------------------------
    ! Start

    srcListSetTabName = stringParameter('srclisttab')
    withMaskSet = booleanParameter('withmaskset')
    if (withMaskSet)&
      maskSetName = stringParameter('maskset')
    selectionExpression = stringParameter('expression')
    templateImageSetName = stringParameter('templateset')

    outImageSetName = stringParameter('outset')
!    withCutoff = booleanParameter('withfluxcut')
!    if (withCutoff)&
!      cutoffFlux = realParameter('fluxcut')
    fluxScale = realParameter('fluxscale')
    halfISize = intParameter('patchhalfsize')

    psfStyle = stringParameter('psfstyle')
    select case(psfStyle)
    case('uniform')
      gaussFrac = realParameter('gaussfrac')
!      width = realParameter('width')
!      sigmaPixels = realParameter('gausswidth')
      sigmaArcsec = realParameter('gausswidth')
!      r0Pixels = realParameter('r0')
      r0Arcsec = realParameter('r0')
!      r0Pixels2 = r0Pixels * r0Pixels
      alpha = realParameter('alpha')
      withGaussianApprox = booleanParameter('withapprox')
      if (withGaussianApprox) then
        gaussApproxStyle = stringParameter('approxstyle')
        select case(gaussApproxStyle)
        case('interpolate')
call error('notYetSupported',"")
        case('subpixels')
          numSubPixelDivs = intParameter('nsubpixels')
        case default
call error('badApproxStyle',"")
        end select
      end if
    case('column')
      widthColName = stringParameter('widthcol')
    case default
call error('badPsfStyle',"")
    end select

!    positionStyle = stringParameter('posstyle')
!    select case(positionStyle)
!    case('radec')
      raColName = stringParameter('racol')
      decColName = stringParameter('deccol')
!    case('pixel')
!      ximaColName = stringParameter('xcol')
!      yimaColName = stringParameter('ycol')
!    case default
!call error('badPositionStyle',"")
!    end select

    rateColName = stringParameter('ratecol')
    tempImageSetName = stringParameter('tempimageset')
    tempSetName = stringParameter('tempsrclistset')
    normalize = booleanParameter('normalize')
    vignet = booleanParameter('vignet')

    withFluxOffset = booleanParameter('withfluxoffset')
    if (withFluxOffset)&
      fluxOffset = realParameter('fluxOffset') ! in cgs/arcsec^2, in the same band as the source fluxes.

    applyAttcalcFudge = booleanParameter('withboresightfudge')
    ! only needed while boresight psi has wrong sign. IMS 2005-11-17

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine getPsfPatch(srcXpixels, srcYpixels, patch)
  !---------------------------------------------------------------------
  ! This generates a PSF model comprising a King function with a gaussian core. There are thus 4 free parameters:
  !
  !	- The King function r_0 and alpha;
  !	- The gaussian sigma;
  !	- The fraction of the total at r == 0 due to the Gaussian.
  !
  ! All these values are supplied as command-line parameters. The default values of r_0 and alpha at time of writing have been set respectively at 5.0 arcsec and 1.5. These were taken from Ghizzardi S, XMM-SOC-CAL-TN-0029 and ..-0022, as representative values of both MOS and PN PSFs, on the optic axis, and at a photon energy of 1 keV (see also TN-0018). The true values vary slightly with energy and off-axis angle (and are not identical for all EPIC cameras) but these differences have been considered unimportant for the present task. IMS 2007-08-07.

!********** this whole task needs to be rearranged for generality and consistency.

    implicit none

    real(single), intent(in) :: srcXpixels, srcYpixels
    real(single), pointer    :: patch(:,:)

!    real(single), parameter :: patchWidthInSigmas = 7.0
    integer(int16) :: xiStart, xiFinis, yiStart, yiFinis, i, xi, yi
    real(single), allocatable :: xs(:), ys(:), xExps(:), yExps(:)
    integer(int16), pointer :: xis(:), yis(:)

    !-------------------------------------------------------------------
    ! Start

    xiStart = lbound(patch,1)
    xiFinis = ubound(patch,1)
    yiStart = lbound(patch,2)
    yiFinis = ubound(patch,2)

    allocate(xs(xiStart:xiFinis), xExps(xiStart:xiFinis))
    allocate(ys(yiStart:yiFinis), yExps(yiStart:yiFinis))
    allocate(xis(xiStart:xiFinis), yis(yiStart:yiFinis))

    xis = (/(i,i=xiStart, xiFinis)/)
    yis = (/(i,i=yiStart, yiFinis)/)

    patch = 0.0
!print *, ''
!print *, xiStart,xiFinis,yiStart,yiFinis, srcXpixels,srcYpixels
!print *, maxval(patch)
    if (gaussFrac > 0.0) then
      if (sigmaPixels <= 0.0) then
        xExps = 0.0
        yExps = 0.0
        where(floor(xis-srcXpixels+0.5)==0) xExps = 1.0
        where(floor(yis-srcYpixels+0.5)==0) yExps = 1.0
!print *, 'maxval(xExps), maxval(yExps)', maxval(xExps), maxval(yExps)
!print *, 'xiStart, xiFinis, srcXpixels:', xiStart, xiFinis, srcXpixels
!print *, 'yiStart, yiFinis, srcYpixels:', yiStart, yiFinis, srcYpixels
!print *, ''

      else
        xs = (real(xis, single) - srcXpixels) / RootTwo / sigmaPixels
        ys = (real(yis, single) - srcYpixels) / RootTwo / sigmaPixels
        xExps = exp(-(xs*xs))
        yExps = exp(-(ys*ys))
      end if

      patch = spread(xExps, 2, (1 + yiFinis - yiStart))&
            * spread(yExps, 1, (1 + xiFinis - xiStart))
!call dumpImageToFits(patch, 'junkpatch.fits')
!print *, maxval(xExps), maxval(yExps), maxval(patch)
!call error('teststop',"")
    end if

    if (gaussFrac < 1.0) then
      ! Add psf wings:
      !
      xs = (real(xis, single) - srcXpixels)
      ys = (real(yis, single) - srcYpixels)
      xs = xs * xs
      ys = ys * ys
      do xi = xiStart, xiFinis
        do yi = yiStart, yiFinis
          patch(xi,yi) = gaussFrac * patch(xi,yi) + (1.0 - gaussFrac)&
            / (1.0 + (xs(xi) + ys(yi)) / r0Pixels2)**alpha
        end do
      end do
    end if

    if (vignet) call vignetPatch(0.7, patch)
!print *, maxval(patch)
    if (normalize) patch = patch / sum(patch)
!print *, maxval(patch)

    deallocate(xis, xs, xExps, yis, ys, yExps)

  end subroutine getPsfPatch

  !---------------------------------------------------------------------
  subroutine calcCubeOfJitteredPsfPatches(patchHyperCube)
  !---------------------------------------------------------------------

    implicit none

!    integer,      intent(in) :: numSubPixels
    real(single), pointer    :: patchHyperCube(:,:,:,:)

    integer(int16) :: xiStart, xiFinis, yiStart, yiFinis, numSubPixels, subXi, subYi
!, i
!    real(single), allocatable :: xs(:), ys(:), xExp(:), yExp(:)
!    integer(int16), pointer :: xis(:), yis(:)
    real(single) :: srcXpixels, srcYpixels
real(single), pointer :: patch(:,:)

    !-------------------------------------------------------------------
    ! Start

    xiStart = lbound(patchHyperCube,1)
    xiFinis = ubound(patchHyperCube,1)
    yiStart = lbound(patchHyperCube,2)
    yiFinis = ubound(patchHyperCube,2)
    numSubPixels = size(patchHyperCube,3)
if (size(patchHyperCube,4) .ne. numSubPixels) call error('',"")
if (lbound(patchHyperCube,3) .ne. 0) call error('',"")
if (lbound(patchHyperCube,4) .ne. 0) call error('',"")

!    allocate(patchHyperCube(xStart:xFinis, yStart:yFinis&
!      , 0:numSubPixels-1, 0:numSubPixels-1))
allocate(patch(xiStart:xiFinis, yiStart:yiFinis))
!    allocate(xs(xStart:xFinis), xExp(xStart:xFinis))
!    allocate(ys(yStart:yFinis), yExp(yStart:yFinis))
!    allocate(xis(xStart:xFinis), yis(yStart:yFinis))

!    xis = (/(i,i=xStart, xFinis)/)
!    yis = (/(i,i=yStart, yFinis)/)
    do subXi = 0, numSubPixels-1
      srcXpixels = (subXi + 0.5) / real(numSubPixels)
!      xs = (real(xis, single) - srcXpixels) / RootTwo / sigmaXpixels
!      xExp = exp(-(xs*xs))
      do subYi = 0, numSubPixels-1
        srcYpixels = (subYi + 0.5) / real(numSubPixels)
!        ys = (real(yis, single) - srcYpixels) / RootTwo / sigmaYpixels
!        yExp = exp(-(ys*ys))

call getPsfPatch(srcXpixels, srcYpixels, patch)

!        patchHyperCube(:,:,subXi,subYi) = spread(xExp, 2, (1 + yFinis - yStart))&
!          * spread(yExp, 1, (1 + xFinis - xStart))
patchHyperCube(:,:,subXi,subYi) = patch

!    call vignetPatch(patch)
      end do
    end do

!    deallocate(xis, xs, xExp, yis, ys, yExp)
deallocate(patch)

  end subroutine calcCubeOfJitteredPsfPatches

  !---------------------------------------------------------------------
  subroutine vignetPatch(vigSkirtFraction, patch)
  !---------------------------------------------------------------------
  ! This is meant to take an array containing a source PSF image and
  ! multiply it by a local vignetting function, so as to avoid the otherwise
  ! sharp drop-off at the edge of the array.
  !
  ! This has now been a bit modified to take account of another problem.
  ! emldetect assumes of course that the psf of the source it is measuring
  ! is of normal shape - at least out as far as the energy-fraction
  ! cutoff specified via the emldetect parameter --ecut. Hence it is no
  ! good, when making these simulated images, to multiply the whole PSF
  ! by some vignetting function, because that will distort its shape. So
  ! what I do now is retain the pristine PSF out to the ecut edge, but
  ! multiply by a vignetting or feathering function outboard of this.
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)    :: vigSkirtFraction
    real(single), intent(inout) :: patch(:,:)

    integer(int16) :: xiSize, yiSize, i, xiPristineHalfSize, yiPristineHalfSize
    real(single) :: xHalfSize, yHalfSize
    real(single), allocatable ::&
      xVector(:), yVector(:), xVigFunction(:), yVigFunction(:)

    !-------------------------------------------------------------------
    ! Start

    xiSize = size(patch, 1)
    yiSize = size(patch, 2)
    xHalfSize = xiSize / 2.0 - 0.5
    yHalfSize = yiSize / 2.0 - 0.5
    xiPristineHalfSize = nint(xHalfSize / (1.0 + vigSkirtFraction))
    yiPristineHalfSize = nint(yHalfSize / (1.0 + vigSkirtFraction))

    allocate(xVector(xiSize), yVector(yiSize))
    allocate(xVigFunction(xiSize), yVigFunction(yiSize))

    xVector = (/(abs(i - (xHalfSize + 1.0)), i=1, xiSize)/)
    ! The +1 prevents patch being 0 all around the rim, which would be
    ! a waste of pixels.

    where (xVector > xiPristineHalfSize)
      xVigFunction = (xVector - xiPristineHalfSize)&
      / ((xHalfSize + 1.0) - xiPristineHalfSize)
      xVigFunction = 1.0 - xVigFunction * xVigFunction
    elsewhere
      xVigFunction = 1.0
    end where

    yVector = (/(abs(i - (yHalfSize + 1.0)), i=1, yiSize)/)
    ! The +1 prevents patch being 0 all around the rim, which would be
    ! a waste of pixels.

    where (yVector > yiPristineHalfSize)
      yVigFunction = (yVector - yiPristineHalfSize)&
      / ((yHalfSize + 1.0) - yiPristineHalfSize)
      yVigFunction = 1.0 - yVigFunction * yVigFunction
    elsewhere
      yVigFunction = 1.0
    end where

    patch = patch * spread(xVigFunction, 2, yiSize)&
                  * spread(yVigFunction, 1, xiSize)

  end subroutine vignetPatch

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module srcmap_mod
